@@ -1,10 +1,4 @@
 
-/* If set, then tries to apply theorem 7 in addition to theorem 5.
- * Normally I would just have this on, but then we'd produce certificates
- * that Math::Prime::Util 0.26 couldn't understand. :(
- */
-#define BLS_THEOREM7 0
-
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -84,6 +78,86 @@
  * these tests on numbers that are very probably prime.
  */
 
+
+static int try_factor(mpz_t f, mpz_t n, int effort)
+{
+  int success = 0;
+
+  if (!success)  success = _GMP_power_factor(n, f);
+
+  if (!success && mpz_cmp_ui(n, (unsigned long)(UV_MAX>>2)) < 0) {
+    UV ui_n = mpz_get_ui(n);
+    UV ui_factors[2];
+    if (!mpz_cmp_ui(n, ui_n)) {
+      success = racing_squfof_factor(ui_n, ui_factors, 200000)-1;
+      if (success)
+        mpz_set_ui(f, ui_factors[0]);
+    }
+  }
+
+  if (effort >= 1) {
+    if (!success)  success = _GMP_pminus1_factor(n, f, 100, 1000);
+    if (!success)  success = _GMP_pminus1_factor(n, f, 1000, 10000);
+  }
+
+  if (!success && effort == 2) {
+    UV log2n = mpz_sizeinbase(n, 2);
+    UV brent_rounds = (log2n <= 64) ? 100000 : 100000 / (log2n-63);
+    int final_B2 = 1000 * (150-(int)log2n);
+    if (log2n < 70) brent_rounds *= 3;
+    if (!success && log2n < 80)  success = _GMP_ECM_FACTOR(n, f, 150, 5);
+    if (!success)  success = _GMP_pbrent_factor(n, f, 3, brent_rounds);
+    if (!success && final_B2 > 10000)  success = _GMP_pminus1_factor(n, f, 10000, final_B2);
+  }
+
+  if (!success && effort >= 3) {
+    if (!success)  success = _GMP_pminus1_factor(n, f, 10000, 200000);
+    if (!success)  success = _GMP_ECM_FACTOR(n, f, 500, 30);
+    if (!success)  success = _GMP_ECM_FACTOR(n, f, 2000, 20);
+  }
+
+  return success;
+}
+static int try_factor2(mpz_t f, mpz_t n, int effort)
+{
+  int success = 0;
+
+  if (!success && effort >= 4) {
+    if (!success)  success = _GMP_pminus1_factor(n, f, 200000, 4000000);
+    if (!success)  success = _GMP_ECM_FACTOR(n, f, 10000, 10);
+  }
+  if (!success && effort >= 5) {
+    UV i;
+    UV ecm_B1 = 10000;
+    UV curves = 10;
+    if (_GMP_is_prob_prime(n)) croak("Internal error in BLS75\n");
+    for (i = 1; i < 18 && !success; i++) {
+      if ((4+i) > (UV)effort) break;
+      ecm_B1 *= 2;
+      success = _GMP_ECM_FACTOR(n, f, ecm_B1, curves);
+    }
+  }
+  return success;
+}
+
+/* F*R = n, F is factored part, R is remainder */
+static void small_factor(mpz_t F, mpz_t R, UV B1)
+{
+  PRIME_ITERATOR(iter);
+  UV tf;
+  for (tf = 2; tf < B1; tf = prime_iterator_next(&iter)) {
+    if (mpz_cmp_ui(R, tf*tf) < 0) break;
+    if (mpz_divisible_ui_p(R, tf)) {
+      do {
+        mpz_mul_ui(F, F, tf);
+        mpz_divexact_ui(R, R, tf);
+      } while (mpz_divisible_ui_p(R, tf));
+    }
+  }
+  prime_iterator_destroy(&iter);
+}
+
+
 /* FIXME:
  *   (1) too much repetitious overhead code in these
  *   (2) way too much copy/paste between Pocklington and BLS
@@ -253,7 +327,7 @@ static int bls_theorem5_limit(mpz_t n, mpz_t A, mpz_t B,
 {
   mpz_mul(t, A, B);
   mpz_add_ui(t, t, 1);
-  if (mpz_cmp(t, n) != 0) croak("BLS75:  A*B != n\n");
+  if (mpz_cmp(t, n) != 0) croak("BLS75:  A*B != n-1\n");
 
   mpz_mul_ui(t, A, 2);
   mpz_tdiv_qr(s, r, B, t);
@@ -269,30 +343,6 @@ static int bls_theorem5_limit(mpz_t n, mpz_t A, mpz_t B,
   return (mpz_cmp(n, y) < 0) ? 1 : 0;
 }
 
-#if BLS_THEOREM7
-static int bls_theorem7_limit(mpz_t n, mpz_t A, mpz_t B, UV B1,
-                              mpz_t t, mpz_t y, mpz_t r, mpz_t s)
-{
-  mpz_mul(t, A, B);
-  mpz_add_ui(t, t, 1);
-  if (mpz_cmp(t, n) != 0) croak("BLS75:  A*B != n-1\n");
-
-  mpz_mul_ui(t, A, 2);
-  mpz_tdiv_qr(s, r, B, t);
-
-  mpz_mul(y, t, A);     /* y = 2*A*A              */
-  mpz_sub_ui(t, r, B1); /* t = r-B1               */
-  mpz_mul(t, t, A);     /* t = A*(r-B1)           */
-  mpz_add(y, y, t);     /* y = 2A^2 + A(r-B1)     */
-  mpz_add_ui(y, y, 1);  /* y = 2A^2 + A(r-B1) + 1 */
-  mpz_mul_ui(t, A, B1); /* t = A*B1               */
-  mpz_add_ui(t, t, 1);  /* t = A*B1+1             */
-  mpz_mul(y, y, t);     /* y = (A*B1+1)*(2A^2+(r-B1)A+1) */
-
-  return (mpz_cmp(n, y) < 0) ? 1 : 0;
-}
-#endif
-
 int _GMP_primality_bls_nm1(mpz_t n, int effort, char** prooftextptr)
 {
   mpz_t nm1, A, B, t, m, f, r, s;
@@ -301,12 +351,7 @@ int _GMP_primality_bls_nm1(mpz_t n, int effort, char** prooftextptr)
   int msp = 0;
   int fsp = 0;
   int success = 1;
-  int theorem7 = 0;
-#if BLS_THEOREM7
-  UV B1 = 20000;
-#else
   UV B1 = 2000;
-#endif
 
   /* We need to do this for BLS */
   if (mpz_even_p(n)) return 0;
@@ -348,15 +393,6 @@ int _GMP_primality_bls_nm1(mpz_t n, int effort, char** prooftextptr)
     if (bls_theorem5_limit(n, A, B, t, m, r, s))
       break;
 
-#if BLS_THEOREM7
-    /* With theorem 7, we can exit even earlier if we know the lower limit
-     * on the size of the factors of B.  We need to do an additional test
-     * at the end (treat B as if it was the last factor). */
-    if (bls_theorem7_limit(n, A, B, B1, t, m, r, s)) {
-      theorem7 = 1;
-      break;
-    }
-#endif
     success = 0;
     /* If the stack is empty, we have failed. */
     if (msp == 0)
@@ -364,96 +400,23 @@ int _GMP_primality_bls_nm1(mpz_t n, int effort, char** prooftextptr)
     /* pop a component off the stack */
     mpz_set(m, mstack[--msp]); mpz_clear(mstack[msp]);
 
-    /* Try to factor it without trying too hard */
-    if (!success)  success = _GMP_power_factor(m, f);
+    success = try_factor(f, m, effort);
 
-    if (!success && mpz_cmp_ui(m, (unsigned long)(UV_MAX>>2)) < 0) {
-      UV ui_m = mpz_get_ui(m);
-      UV ui_factors[2];
-      if (!mpz_cmp_ui(m, ui_m)) {
-        success = racing_squfof_factor(ui_m, ui_factors, 200000)-1;
-        if (success)
-          mpz_set_ui(f, ui_factors[0]);
-      }
-    }
-    if (effort >= 1) {
-      if (!success)  success = _GMP_pminus1_factor(m, f, 100, 1000);
-      if (!success)  success = _GMP_pminus1_factor(m, f, 1000, 10000);
-    }
-    if (!success && effort == 2) {
-      /* These keep the time in the realm of 2ms per input number, which
-       * means they adjust for size.  Success varies by size:
-       *   99% of  70-bit inputs
-       *   90% of  90-bit inputs
-       *   58% of 120-bit inputs
-       *   25% of 160-bit inputs
-       *   10% of 190-bit inputs
-       */
-      UV log2m = mpz_sizeinbase(m, 2);
-      UV brent_rounds = (log2m <= 64) ? 100000 : 100000 / (log2m-63);
-      int final_B2 = 1000 * (150-(int)log2m);
-      if (log2m < 70) brent_rounds *= 3;
-      if (!success && log2m < 80)  success = _GMP_ECM_FACTOR(m, f, 150, 5);
-      if (!success)  success = _GMP_pbrent_factor(m, f, 3, brent_rounds);
-      if (!success && final_B2 > 10000)  success = _GMP_pminus1_factor(m, f, 10000, final_B2);
-    }
-    if (!success && effort >= 3) {
-      if (!success)  success = _GMP_pminus1_factor(m, f, 10000, 200000);
-      if (!success)  success = _GMP_ECM_FACTOR(m, f, 500, 30);
-      if (!success)  success = _GMP_ECM_FACTOR(m, f, 2000, 20);
-    }
     if (!success && effort >= 5) {
       /* The factor isn't trivial, but our size is relatively small.  QS. */
       if (!success && mpz_sizeinbase(m,10)>=30 && mpz_sizeinbase(m,10)<55) {
         INNER_QS_FACTOR(m, _GMP_primality_bls_nm1);
       }
     }
-    if (!success && effort >= 4) {
-      if (!success)  success = _GMP_pminus1_factor(m, f, 200000, 4000000);
-      if (!success)  success = _GMP_ECM_FACTOR(m, f, 10000, 10);
-    }
-#if BLS_THEOREM7
-    /* Could we exit now if B1 was larger? */
-    if (!success && effort > 2) {
-      UV newB1 = B1;
-      if      (bls_theorem7_limit(n,A,B,    10000,t,f,r,s)) newB1 = 10000;
-      else if (bls_theorem7_limit(n,A,B,  1000000,t,f,r,s)) newB1 = 1000000;
-      else if (bls_theorem7_limit(n,A,B, 10000000,t,f,r,s)) newB1 = 10000000;
-      if (newB1 > B1) {
-        UV tf = _GMP_trial_factor(m, B1, newB1);
-        if (tf == 0) {
-          /* Put m back, restart */
-          B1 = newB1;
-          mpz_init_set(mstack[msp++], m);
-          success = 1;
-          continue; /* Go back to the top and we'll retest with this B1 */
-        } else {
-          /* Holy trial factorization, Batman! */
-          B1 = tf;
-          mpz_set_ui(f, tf);
-          success = 1;
-        }
-      }
-    }
-#endif
     if (!success && effort > 5) {  /* do here only if effort > 5 */
       /* QS.  Uses lots of memory, but finds multiple factors quickly */
       if (!success && mpz_sizeinbase(m,10)>=30 && mpz_sizeinbase(m,2)<270) {
         INNER_QS_FACTOR(m, _GMP_primality_bls_nm1);
       }
     }
-    if (!success && effort >= 5) {
-      UV i;
-      UV ecm_B1 = 10000;
-      UV curves = 10;
-      if (_GMP_is_prob_prime(m)) croak("Internal error in BLS75\n");
-      for (i = 1; i < 18; i++) {
-        if ((4+i) > (UV)effort) break;
-        ecm_B1 *= 2;
-        success = _GMP_ECM_FACTOR(m, f, ecm_B1, curves);
-        if (success) break;
-      }
-    }
+    if (!success)
+      success = try_factor2(f, m, effort);
+
     /* If we couldn't factor m and the stack is empty, we've failed. */
     if ( (!success) && (msp == 0) )
       break;
@@ -469,17 +432,45 @@ int _GMP_primality_bls_nm1(mpz_t n, int effort, char** prooftextptr)
     mpz_clear(mstack[msp]);
   msp = 0;
 
-  if (success > 0) {
-    /* Calculate r and s from final A/B values */
-    mpz_mul_ui(t, A, 2);
-    mpz_tdiv_qr(s, r, B, t);
+  /* Sort factors found from largest to smallest, but 2 must be at start. */
+  {
+    int i, j;
+    for (i = 2; i < fsp; i++) {
+      for (j = i; j > 1 && mpz_cmp(fstack[j-1], fstack[j]) < 0; j--) {
+        mpz_set(t, fstack[j-1]);
+        mpz_set(fstack[j-1], fstack[j]);
+        mpz_set(fstack[j], t);
+      }
+    }
+  }
 
-    mpz_mul(t, r, r);
-    mpz_submul_ui(t, s, 8);   /* t = r^2 - 8s */
-    /* N is prime if and only if s=0 OR t not a perfect square */
-    /* So, N is a composite if s != 0 and t is perfect square */
-    if (mpz_sgn(s) != 0 && mpz_perfect_square_p(t))
-      success = -1;
+  /* Shrink to smallest set and verify conditions. */
+  if (success > 0) {
+    int i;
+    mpz_set_ui(A, 1);
+    mpz_set(B, nm1);
+    for (i = 0; i < fsp; i++) {
+      if (bls_theorem5_limit(n, A, B, t, m, r, s))
+        break;
+      do {
+        mpz_mul(A, A, fstack[i]);
+        mpz_divexact(B, B, fstack[i]);
+      } while (mpz_divisible_p(B, fstack[i]));
+    }
+    /* Delete any extra factors */
+    while (i < fsp)
+      mpz_clear(fstack[--fsp]);
+    /* Verify Q[0] = 2 */
+    if (mpz_cmp_ui(fstack[0], 2) != 0)
+      croak("Internal error: 2 not at start of fstack");
+    /* Verify conditions */
+    success = 0;
+    if (bls_theorem5_limit(n, A, B, t, m, r, s)) {
+      mpz_mul(t, r, r);
+      mpz_submul_ui(t, s, 8);   /* t = r^2 - 8s */
+      /* N is prime if and only if s=0 OR t not a perfect square */
+      success = (mpz_sgn(s) == 0 || !mpz_perfect_square_p(t))  ?  1  :  -1;
+    }
   }
 
   if (success > 0) {
@@ -490,12 +481,6 @@ int _GMP_primality_bls_nm1(mpz_t n, int effort, char** prooftextptr)
     mpz_init(p);
     mpz_init(ap);
 
-    if (theorem7) {
-      /* Theorem 7 used, so add B to the factor list so it gets checked */
-      if (fsp >= PRIM_STACK_SIZE) { croak("BLS75 stack overflow\n"); }
-      mpz_init_set(fstack[fsp++], B);
-    }
-
     for (pcount = 0; success && pcount < fsp; pcount++) {
       PRIME_ITERATOR(iter);
       mpz_set(p, fstack[pcount]);
@@ -526,36 +511,30 @@ int _GMP_primality_bls_nm1(mpz_t n, int effort, char** prooftextptr)
     mpz_clear(ap);
   }
   if (success > 0 && prooftextptr != 0) {
-    int i, prooflen;
+    int i;
     char *proofstr, *proofptr;
+    int curprooflen = (*prooftextptr == 0) ? 0 : strlen(*prooftextptr);
+    int myprooflen = (5 + mpz_sizeinbase(n, 10)) * (2 + fsp + msp) + 200;
+
     if (fsp != msp) croak("Different f and a counts\n");
-    prooflen = mpz_sizeinbase(n, 10) * (1 + fsp + msp) + 200;
-    New(0, proofstr, prooflen, char);
+    New(0, proofstr, myprooflen + curprooflen + 1, char);
     proofptr = proofstr;
-    *proofptr = 0;
-    proofptr += gmp_sprintf(proofptr, "%Zd :", n);
-    if (theorem7) {
-      proofptr += gmp_sprintf(proofptr, " N-1 T7 :");
-      proofptr += gmp_sprintf(proofptr, " %lu %Zd %Zd :",
-                       (unsigned long)B1, fstack[--fsp], mstack[--msp]);
-       mpz_clear(fstack[fsp]);  mpz_clear(mstack[msp]);
-    } else {
-      proofptr += gmp_sprintf(proofptr, " N-1 T5 :");
-    }
-    for (i = 0; i < fsp; i++)
-      proofptr += gmp_sprintf(proofptr, " %Zd", fstack[i]);
-    proofptr += gmp_sprintf(proofptr, " :");
+    proofptr += gmp_sprintf(proofptr, "Type BLS5\nN  %Zd\n", n);
+    /* Q[0] is always 2 */
+    for (i = 1; i < fsp; i++)
+      proofptr += gmp_sprintf(proofptr, "Q[%d]  %Zd\n", i, fstack[i]);
+    /* A[i] only printed if not 2 */
     for (i = 0; i < msp; i++)
-      proofptr += gmp_sprintf(proofptr, " %Zd", mstack[i]);
-    proofptr += gmp_sprintf(proofptr, "\n");
-    /* Set or append */
-    if (*prooftextptr == 0) {
-      *prooftextptr = proofstr;
-    } else {
-      Renew(*prooftextptr, strlen(*prooftextptr) + strlen(proofstr) + 1, char);
-      (void) strcat(*prooftextptr, proofstr);
-      Safefree(proofstr);
+      if (mpz_cmp_ui(mstack[i], 2) != 0)
+        proofptr += gmp_sprintf(proofptr, "A[%d]  %Zd\n", i, mstack[i]);
+    proofptr += gmp_sprintf(proofptr, "----\n");
+    /* Set or prepend */
+    if (*prooftextptr) {
+      proofptr += gmp_sprintf(proofptr, "\n");
+      strcat(proofptr, *prooftextptr);
+      Safefree(*prooftextptr);
     }
+    *prooftextptr = proofstr;
   }
   while (fsp-- > 0)
     mpz_clear(fstack[fsp]);
@@ -573,3 +552,224 @@ int _GMP_primality_bls_nm1(mpz_t n, int effort, char** prooftextptr)
   if (success > 0) return 2;
   return 1;
 }
+
+
+
+/* Given an n where we're factored n-1 down to p, check BLS theorem 3 */
+int _GMP_primality_bls_3(mpz_t n, mpz_t p, UV* reta)
+{
+  mpz_t nm1, m, t, t2;
+  int rval = 0;
+
+  if (reta) *reta = 0;
+  if (mpz_cmp_ui(n, 2) <= 0 || mpz_even_p(n) || mpz_even_p(p))
+    return 0;                 /* n is <= 2, n is even, or p is even */
+  if (!_GMP_is_prob_prime(p))
+    return 0;                 /* p is not a probable prime */
+
+  mpz_init(nm1);  mpz_init(m);  mpz_init(t);  mpz_init(t2);
+  mpz_sub_ui(nm1, n, 1);
+  mpz_divexact(m, nm1, p);
+  mpz_mul(t, m, p);
+  if (mpz_cmp(nm1, t) != 0)
+    goto end_bls3;           /* m*p != n+1 */
+
+  mpz_mul_ui(t, p, 2);
+  mpz_add_ui(t, t, 1);
+  mpz_sqrt(t2, n);
+  if (mpz_cmp(t, t2) <= 0)
+    goto end_bls3;           /* 2p+1 <= sqrt(n) */
+
+  {
+    /* N-1 = mp, p is an odd probable prime, and 2p+1 > sqrt(n).
+     * Now find an 'a' where a^(n-1)/2 = -1 mod n, a^(m/2) != -1 mod n. */
+    PRIME_ITERATOR(iter);
+    UV const alimit = 1000;
+    UV a;
+    for (a = 2; a <= alimit; a = prime_iterator_next(&iter)) {
+      mpz_set_ui(t2, a);
+      mpz_divexact_ui(t, m, 2);
+      mpz_powm(t, t2, t, n);       /* a^(m/2) mod n */
+      if (mpz_cmp(t, nm1) == 0)
+        continue;
+      mpz_divexact_ui(t, nm1, 2);
+      mpz_powm(t, t2, t, n);       /* a^((n-1)/2) mod n */
+      if (mpz_cmp(t, nm1) != 0)
+        continue;
+      rval = 2;
+      if (reta) *reta = a;
+      break;
+    }
+    prime_iterator_destroy(&iter);
+  }
+
+end_bls3:
+  mpz_clear(nm1);  mpz_clear(m);  mpz_clear(t);  mpz_clear(t2);
+  return rval;
+}
+
+/* Given an n where we're factored n+1 down to f, check BLS theorem 15 */
+int _GMP_primality_bls_15(mpz_t n, mpz_t f, IV* lp, IV* lq)
+{
+  mpz_t np1, m, t, t2;
+  int rval = 0;
+
+  if (lp) *lp = 0;
+  if (lq) *lq = 0;
+  if (mpz_cmp_ui(n, 2) <= 0 || mpz_even_p(n) || mpz_even_p(f))
+    return 0;                 /* n is <= 2, n is even, or f is even */
+  if (!_GMP_is_prob_prime(f))
+    return 0;                 /* f is not a probable prime */
+
+  mpz_init(np1);  mpz_init(m);  mpz_init(t);  mpz_init(t2);
+  mpz_add_ui(np1, n, 1);
+  mpz_divexact(m, np1, f);
+  mpz_mul(t, m, f);
+  if (mpz_cmp(np1, t) != 0)
+    goto end_bls15;           /* m*f != n+1 */
+
+  mpz_mul_ui(t, f, 2);
+  mpz_sub_ui(t, t, 1);
+  mpz_sqrt(t2, n);
+  if (mpz_cmp(t, t2) <= 0)
+    goto end_bls15;           /* 2f-1 <= sqrt(n) */
+
+  {
+    /* N+1 = mf, f is an odd probable prime, and 2f-1 > sqrt(n).
+     * Now find a Lucas sequence V_k with discriminant D s.t. D/N = -1
+     * where N divides V_(N+1)/2 and N does not divide V_m/2. */
+    IV d, p, q;
+    mpz_t U, V, k;
+
+    q = 2;
+    mpz_init(U);  mpz_init(V);  mpz_init(k);
+
+    /* Primo gave me the idea of this p/q selection method */
+    for (q = 2; q < 1000; q++) {
+      p = (q % 2) ? 2 : 1;
+      d = p*p - 4*q;
+      mpz_set_si(t, d);
+      if (mpz_jacobi(t, n) != -1)
+        continue;
+      /* we have a d/p/q where d = -1.  Check the Lucas sequences. */
+      mpz_divexact_ui(k, m, 2);
+      _GMP_lucas_seq(U, V, n, p, q, k,    t, t2);
+      if (mpz_sgn(V) != 0) {
+        mpz_divexact_ui(k, np1, 2);
+        _GMP_lucas_seq(U, V, n, p, q, k,    t, t2);
+        if (mpz_sgn(V) == 0) {
+          rval = 2;
+          if (lp) *lp = p;
+          if (lq) *lq = q;
+          break;
+        }
+      }
+    }
+    mpz_clear(U);  mpz_clear(V);  mpz_clear(k);
+  }
+
+end_bls15:
+  mpz_clear(np1);  mpz_clear(m);  mpz_clear(t);  mpz_clear(t2);
+  return rval;
+}
+
+/* Given an n, try using BLS75 theorem 15, N+1 = mq.
+ * Note: this does _not_ prove n is prime!  If it returns 1, then we have
+ * found a q/D that satisfy theorem 15, but we leave proving q for the caller.
+ */
+int _GMP_primality_bls_np1_split(mpz_t n, int effort, mpz_t q, IV* lp, IV* lq)
+{
+  mpz_t np1, m, f, sqrtn, t;
+  int success = 1;
+  UV B1 = 2000;
+
+  /* We need to do this for BLS */
+  if (mpz_even_p(n)) return 0;
+
+  mpz_init(np1);  mpz_init(m);  mpz_init(f);  mpz_init(sqrtn);  mpz_init(t);
+  mpz_add_ui(np1, n, 1);
+  mpz_set_ui(m, 1);
+  mpz_set(q, np1);
+  mpz_sqrt(sqrtn, n);
+
+  small_factor(m, q, B1);
+
+  while (success) {
+    success = 0;
+    mpz_mul_ui(t, q, 2);
+    mpz_sub_ui(t, t, 1);
+    if (mpz_cmp(t, sqrtn) <= 0)
+      break;
+    if (_GMP_is_prob_prime(q)) {
+      success = 1;
+      break;
+    }
+    success = try_factor(f, q, effort);
+    if (!success)
+      success = try_factor2(f, q, effort);
+    if (success) {
+      mpz_divexact(q, q, f);
+      if (mpz_cmp(q, f) < 0) { mpz_set(t, q); mpz_set(q, f); mpz_set(f, t); }
+      mpz_mul(m, m, f);
+    }
+  }
+
+  if (success)
+    success = _GMP_primality_bls_15(n, q, lp, lq);
+
+  mpz_clear(np1);
+  mpz_clear(m);
+  mpz_clear(f);
+  mpz_clear(sqrtn);
+  mpz_clear(t);
+  return success;
+}
+
+/* Given an n, try using BLS75 theorem 3, N-1 = mp. */
+int _GMP_primality_bls_nm1_split(mpz_t n, int effort, mpz_t p, UV *reta)
+{
+  mpz_t nm1, m, f, sqrtn, t;
+  int success = 1;
+  UV B1 = 2000;
+
+  /* We need to do this for BLS */
+  if (mpz_even_p(n)) return 0;
+
+  mpz_init(nm1);  mpz_init(m);  mpz_init(f);  mpz_init(sqrtn);  mpz_init(t);
+  mpz_sub_ui(nm1, n, 1);
+  mpz_set_ui(m, 1);
+  mpz_set(p, nm1);
+  mpz_sqrt(sqrtn, n);
+
+  small_factor(m, p, B1);
+
+  while (success) {
+    success = 0;
+    mpz_mul_ui(t, p, 2);
+    mpz_add_ui(t, t, 1);
+    if (mpz_cmp(t, sqrtn) <= 0)
+      break;
+    if (_GMP_is_prob_prime(p)) {
+      success = 1;
+      break;
+    }
+    success = try_factor(f, p, effort);
+    if (!success)
+      success = try_factor2(f, p, effort);
+    if (success) {
+      mpz_divexact(p, p, f);
+      if (mpz_cmp(p, f) < 0) { mpz_set(t, p); mpz_set(p, f); mpz_set(f, t); }
+      mpz_mul(m, m, f);
+    }
+  }
+
+  if (success)
+    success = _GMP_primality_bls_3(n, p, reta);
+
+  mpz_clear(nm1);
+  mpz_clear(m);
+  mpz_clear(f);
+  mpz_clear(sqrtn);
+  mpz_clear(t);
+  return success;
+}
